



qserver --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
start-re-manager --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
qserver-list-plans-devices --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
qserver-zmq-keys --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
qserver-console-monitor --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -c "import os; import bluesky_queueserver; assert os.path.isdir(os.path.join(os.path.dirname(bluesky_queueserver.__file__), 'profile_collection_sim'))"
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
